unit necklace;

interface

procedure init;
procedure create (from : longint; operation : char; on_left : boolean; param : longint);
function pearl (neck_id : longint; on_left : boolean) : longint;

implementation

type neckState = (EMPTY, EVALUATED, THUNK_MOVE, THUNK_TAKE);
     pneckStack = ^neckStack;
     neckStack = record
                   case state : neckState of
		     EMPTY:   ( { nothing here } );
	             EVALUATED: ( head : longint; tail : pneckStack; );
		     THUNK_MOVE:   { equal to toAppendF ++ reverse (drop toSkip toReverse) ++ toAppendR }
		       (toReverse, toAppendF, toAppendR : pneckStack;
		       toSkip : longint;);
		     THUNK_TAKE:   { equal to take toTake contents }
		       (contents : pneckStack; toTake : longint;);
		 end;


     {invariants :

        frontL <= 3 * rearL + 1
	rearL <= 3 * frontL + 1

	We start to balance when rearL = X and frontL >= 3 * X + 2;
	   (or symmetrically on the other side).  Let
	   C = 3 * X + 2 - frontL, C <= 2.  At that moment,
	   evalF and evalR (unevaluated parts of front and rear)
	   are guaranteed to be fully evaluated.

	We rebalance the stacks in the following way:

	frontL = 2 * X + 1
	rearL = 2 * X + 1 + C
	evalF = front = take (2 * X + 1) old_front
	evalR = rear = oldRear ++ reverse (drop (2 * X + 1) old_front) ++ []

        To get front evaluated, (2 * X + 1) steps are needed.
	To get rear evaluated, (2 * X + 1) steps are needed for drop,
	   X + C + 1 for reverse and append at the end and X for append at the
	   front.  Front append and drop are done in parallel, with drop going
	   twice as fast; in total we need (2 * X + C + 1) steps.

	At least 4  * (X + 1) / 3 operations must be performed before next rebalance.
	At least X operations before the reversed part of front is accessed; at that
	moment, we must have the reversed part evaluated.  To achieve this, we perform
	initially three rebalancing operations and then two rebalancing operations for
	each queue operation.
	}

     neck = record
              front, rear, evalF, evalR : pneckStack;
	      frontL, rearL : longint;
	    end;

function newEvaluated (h : longint; t : pneckStack) : pneckStack;
var nw : pneckStack;
begin

 new (nw);
 nw^.state := EVALUATED;
 nw^.head := h;
 nw^.tail := t;
 newEvaluated := nw;
end;

procedure evalStep (var s : pneckStack);
var fa, dr, fe, nw : pneckStack;
    sk : longint;
begin
  case s^.state of
    EMPTY: ;

    EVALUATED:
      s := s^.tail;

    THUNK_MOVE:
      begin
      fa := s^.toAppendF;
      dr := s^.toReverse;
      fe := s^.toAppendR;
      sk := s^.toSkip;

      if sk > 1 then
        begin
	  new(nw);
	  s^.state := EVALUATED;
	  s^.head := fa^.head;
	  s^.tail := nw;

	  assert (fa^.state = EVALUATED);
	  fa := fa^.tail;
	  assert (dr^.state = EVALUATED);
	  dr := dr^.tail;
	  assert (dr^.state = EVALUATED);
	  dr := dr^.tail;

	  nw^.state := THUNK_MOVE;
	  nw^.toAppendF := fa;
	  nw^.toReverse := dr;
	  nw^.toAppendR := fe;
	  nw^.toSkip := sk - 2;
          s := nw;
	  exit;
	end;

       if sk = 1 then
          begin
	    s^.toSkip := 0;
	    assert (dr^.state = EVALUATED);
	    dr := dr^.tail;
	    assert (fa^.state = EMPTY);
	  end;

        assert (dr^.state = EVALUATED);
	if dr^.tail^.state = EMPTY then
	  begin
	    s^.state := EVALUATED;
	    s^.head := dr^.head;
	    s^.tail := fe;
	    s := dr^.tail;
	    exit;
	  end;

	s^.toReverse := dr^.tail;
	s^.toAppendR := newEvaluated (dr^.head, fe);
      end;
    
    THUNK_TAKE:
      begin
      fe := s^.contents;
      assert (fe^.state = EVALUATED);
      sk := s^.toTake;

      new (nw);
      s^.state := EVALUATED;
      s^.head := fe^.head;
      s^.tail := nw;

      if sk = 1 then
         nw^.state := EMPTY
      else
         begin
	   nw^.state := THUNK_TAKE;
	   nw^.toTake := sk - 1;
	   nw^.contents := fe^.tail;
	 end;
      s := nw;
      end
  end;
end;

procedure balanceStep (var n : neck);
begin
  evalStep (n.evalF);
  evalStep (n.evalR);
end;

function fullyEvaluated (p:pneckStack) : boolean;
begin
  while p^.state <> EMPTY do
    begin
      if p^.state <> EVALUATED then
        begin
	  fullyEvaluated := false;
	  exit;
	end;
      p := p^.tail;
    end;

  fullyEvaluated := true;
end;

procedure maybe_balance (var n : neck);
var x, c : longint;
begin
  if n.frontL >= 3 * n.rearL + 2 then
    begin
      x := n.rearL;
      c := n.frontL - (3 * x + 2);
{      assert (fullyEvaluated (n.evalF));
      assert (fullyEvaluated (n.evalR));}
      new (n.evalF);
      n.evalF^.state := THUNK_TAKE;
      n.evalF^.toTake := 2 * x + 1;
      n.evalF^.contents := n.front;

      new (n.evalR);
      n.evalR^.state := THUNK_MOVE;
      n.evalR^.toAppendF := n.rear;
      n.evalR^.toReverse := n.front;
      n.evalR^.toSkip := 2 * x + 1;
      new (n.evalR^.toAppendR);
      n.evalR^.toAppendR^.state := EMPTY;

      n.frontL := 2 * x + 1;
      n.rearL := 2 * x + 1 + c;
    end
  else if n.rearL >= 3 * n.frontL + 2 then
    begin
      x := n.frontL;
      c := n.rearL - (3 * x + 2);

{      assert (fullyEvaluated (n.evalF));
      assert (fullyEvaluated (n.evalR));}
      new (n.evalR);
      n.evalR^.state := THUNK_TAKE;
      n.evalR^.toTake := 2 * x + 1;
      n.evalR^.contents := n.rear;

      new (n.evalF);
      n.evalF^.state := THUNK_MOVE;
      n.evalF^.toAppendF := n.front;
      n.evalF^.toReverse := n.rear;
      n.evalF^.toSkip := 2 * x + 1;
      new (n.evalF^.toAppendR);
      n.evalF^.toAppendR^.state := EMPTY;

      n.rearL := 2 * x + 1;
      n.frontL := 2 * x + 1 + c;
    end
  else
    exit;

  n.front := n.evalF;
  n.rear := n.evalR;
  balanceStep (n);
  balanceStep (n);
  balanceStep (n);
end;

procedure pushF (n : neck; v : longint; var nw : neck);
begin
  nw := n;
  balanceStep (nw);
  balanceStep (nw);
  inc (nw.frontL);
  nw.front := newEvaluated (v, nw.front);
  maybe_balance (nw);
end;

procedure pushR (n : neck; v : longint; var nw : neck);
begin
  nw := n;
  balanceStep (nw);
  balanceStep (nw);
  inc (nw.rearL);
  nw.rear := newEvaluated (v, nw.rear);
  maybe_balance (nw);
end;

procedure popR (n : neck; var nw : neck); forward;

procedure popF (n : neck; var nw : neck);
begin
  if n.frontL = 0 then
    begin
      assert (n.front^.state = EMPTY);
      assert (n.rearL = 1);
      popR (n, nw);
      exit;
    end;
  nw := n;
  balanceStep (nw);
  balanceStep (nw);

  dec (nw.frontL);
  assert (nw.front^.state = EVALUATED);
  nw.front := nw.front^.tail;
  maybe_balance (nw);
end;

procedure popR (n : neck; var nw : neck);
begin
  if n.rearL = 0 then
    begin
      assert (n.rear^.state = EMPTY);
      assert (n.frontL = 1);
      popF (n, nw);
      exit;
    end;
  nw := n;
  balanceStep (nw);
  balanceStep (nw);

  dec (nw.rearL);
  assert (nw.rear^.state = EVALUATED);
  nw.rear := nw.rear^.tail;
  maybe_balance (nw);
end;

function topF (n : neck) : longint;
begin
  if n.frontL = 0 then
    begin
      assert (n.front^.state = EMPTY);
      assert (n.rearL = 1);
      assert (n.rear^.state = EVALUATED);
      topF := n.rear^.head;
    end
  else
    begin
      assert (n.front^.state = EVALUATED);
      topF := n.front^.head;
    end;
end;

function topR (n : neck) : longint;
begin
  if n.rearL = 0 then
    begin
      assert (n.rear^.state = EMPTY);
      assert (n.frontL = 1);
      assert (n.front^.state = EVALUATED);
      topR := n.front^.head;
    end
  else
    begin
      assert (n.rear^.state = EVALUATED);
      topR := n.rear^.head;
    end;
end;

procedure empty_neck (var nck: neck);
begin
 new (nck.front);
 nck.front^.state := EMPTY;
 new (nck.rear);
 nck.rear^.state := EMPTY;
 new (nck.evalF);
 nck.evalF^.state := EMPTY;
 new (nck.evalR);
 nck.evalR^.state := EMPTY;
 nck.frontL := 0;
 nck.rearL := 0;
end;

var necks:array[0 .. 1000000] of neck;
    nnecks : longint;

procedure init;
begin
 nnecks := 1;
 empty_neck (necks[0]);
end;

procedure create (from : longint; operation : char; on_left : boolean; param : longint);
begin
  if operation = 'A' then
    begin
      if on_left then pushF (necks[from], param, necks[nnecks])
                 else pushR (necks[from], param, necks[nnecks]);
    end
  else
    begin
      if on_left then popF (necks[from], necks[nnecks])
                 else popR (necks[from], necks[nnecks]);
    end;
  inc (nnecks);
end;

function pearl (neck_id : longint; on_left : boolean) : longint;
begin
  if on_left then
    pearl := topF (necks[neck_id])
  else
    pearl := topR (necks[neck_id]);
end;

procedure dump_stack (t : pneckStack);
begin
  case t^.state of
    EMPTY: write ('[]');
    EVALUATED: begin
               write (t^.head, ' : ');
	       dump_stack (t^.tail);
	       end;
    THUNK_MOVE: begin
                write ('(');
		dump_stack (t^.toAppendF);
                write (') ++ reverse (drop ', t^.toSkip, ' (');
		dump_stack (t^.toReverse);
                write (') ++ ');
		dump_stack (t^.toAppendR);
		end;
   THUNK_TAKE: begin
   	       write ('take ', t^.toTake, ' ');
	       dump_stack (t^.contents);
	       end;
   end;
end;

procedure dump_necklace(nck : neck);
begin
  write ('  front (', nck.frontL, '): ');
  dump_stack (nck.front);
  writeln;
  write ('  rear (', nck.rearL, '): ');
  dump_stack (nck.rear);
  writeln;
  write ('  evalF: ');
  dump_stack (nck.evalF);
  writeln;
  write ('  evalR: ');
  dump_stack (nck.evalR);
  writeln;
end;

procedure dump_necklaces;
var i : longint;
begin
  for i := 0 to nnecks - 1 do
    begin
    writeln ('Necklace ', i);
    dump_necklace (necks[i]);
    end;
end;

end.

